from supervisely.nn.benchmark.utils.semantic_segmentation.evaluator import Evaluator
from supervisely.nn.benchmark.utils.semantic_segmentation.loader import (
    build_segmentation_loader,
)


def calculate_metrics(
    gt_dir,
    pred_dir,
    boundary_width,
    boundary_iou_d,
    num_workers,
    class_names,
    result_dir,
    progress=None,
):
    if boundary_width % 1 == 0:
        boundary_width = int(boundary_width)
    evaluator = Evaluator(
        class_names=class_names,
        boundary_width=boundary_width,
        boundary_implementation="exact",
        boundary_iou_d=boundary_iou_d,
        result_dir=result_dir,
        progress=progress,
    )
    loader = build_segmentation_loader(
        pred_dir=pred_dir,
        gt_dir=gt_dir,
        gt_label_map=None,
        pred_label_map=None,
        num_workers=num_workers,
    )
    eval_data = evaluator.evaluate(loader)
    return eval_data



from typing import List

import supervisely.nn.benchmark.comparison.detection_visualization.text_templates as texts
from supervisely.nn.benchmark.comparison.base_visualizer import BaseComparisonVisualizer
from supervisely.nn.benchmark.comparison.detection_visualization.vis_metrics import (
    AveragePrecisionByClass,
    CalibrationScore,
    ExplorePredictions,
    LocalizationAccuracyIoU,
    OutcomeCounts,
    Overview,
    PrCurve,
    PrecisionRecallF1,
    Speedtest,
)
from supervisely.nn.benchmark.object_detection.evaluator import (
    ObjectDetectionEvalResult,
)
from supervisely.nn.benchmark.visualization.widgets import (
    ContainerWidget,
    GalleryWidget,
    MarkdownWidget,
    SidebarWidget,
)


class DetectionComparisonVisualizer(BaseComparisonVisualizer):
    vis_texts = texts
    ann_opacity = 0.5

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.eval_results: List[ObjectDetectionEvalResult]

    def _create_widgets(self):
        # Modal Gellery
        self.diff_modal_table = self._create_diff_modal_table()
        self.explore_modal_table = self._create_explore_modal_table(self.diff_modal_table.id)

        # Notifcation
        self.clickable_label = self._create_clickable_label()

        # Speedtest init here for overview
        speedtest = Speedtest(self.vis_texts, self.comparison.eval_results)

        # Overview
        overview = Overview(self.vis_texts, self.comparison.eval_results)
        overview.team_id = self.comparison.team_id
        self.header = self._create_header()
        self.overviews = self._create_overviews(overview)
        self.overview_md = overview.overview_md
        self.key_metrics_md = self._create_key_metrics()
        self.key_metrics_table = overview.get_table_widget(speedtest.latency, speedtest.fps)
        self.overview_chart = overview.chart_widget
        self.iou_per_class_thresholds_md = overview.not_matched_iou_per_class_thresholds_md

        columns_number = len(self.comparison.eval_results) + 1  # +1 for GT
        self.explore_predictions_modal_gallery = self._create_explore_modal_table(columns_number)
        explore_predictions = ExplorePredictions(
            self.vis_texts,
            self.comparison.eval_results,
            explore_modal_table=self.explore_predictions_modal_gallery,
        )
        self.explore_predictions_md = explore_predictions.difference_predictions_md
        self.explore_predictions_gallery = explore_predictions.explore_gallery

        # Outcome Counts
        outcome_counts = OutcomeCounts(
            self.vis_texts,
            self.comparison.eval_results,
            explore_modal_table=self.explore_modal_table,
        )
        self.outcome_counts_md = self._create_outcome_counts_md()
        self.outcome_counts_diff_md = self._create_outcome_counts_diff_md()
        self.outcome_counts_main = outcome_counts.chart_widget_main
        self.outcome_counts_comparison = outcome_counts.chart_widget_comparison

        # Precision-Recall Curve
        pr_curve = PrCurve(self.vis_texts, self.comparison.eval_results)
        self.pr_curve_md = pr_curve.markdown_widget
        self.pr_curve_collapsed_widgets = pr_curve.collapsed_widget
        self.pr_curve_table = pr_curve.table_widget
        self.pr_curve_chart = pr_curve.chart_widget

        # Average Precision by Class
        avg_prec_by_class = AveragePrecisionByClass(
            self.vis_texts,
            self.comparison.eval_results,
            explore_modal_table=self.explore_modal_table,
        )
        self.avg_prec_by_class_md = avg_prec_by_class.markdown_widget
        self.avg_prec_by_class_chart = avg_prec_by_class.chart_widget

        # Precision, Recall, F1
        precision_recall_f1 = PrecisionRecallF1(
            self.vis_texts,
            self.comparison.eval_results,
            explore_modal_table=self.explore_modal_table,
        )
        self.precision_recall_f1_md = precision_recall_f1.markdown_widget
        self.precision_recall_f1_table = precision_recall_f1.table_widget
        self.precision_recall_f1_chart = precision_recall_f1.chart_main_widget
        self.precision_per_class_title_md = precision_recall_f1.precision_per_class_title_md
        self.precision_per_class_chart = precision_recall_f1.chart_precision_per_class_widget
        self.recall_per_class_title_md = precision_recall_f1.recall_per_class_title_md
        self.recall_per_class_chart = precision_recall_f1.chart_recall_per_class_widget
        self.f1_per_class_chart = precision_recall_f1.chart_f1_per_class_widget
        self.f1_per_class_title_md = precision_recall_f1.f1_per_class_title_md

        # Classification Accuracy
        # TODO: ???

        # Localization Accuracy (IoU)
        loc_acc = LocalizationAccuracyIoU(self.vis_texts, self.comparison.eval_results)
        self.loc_acc_header_md = loc_acc.header_md
        self.loc_acc_iou_distribution_md = loc_acc.iou_distribution_md
        self.loc_acc_chart = loc_acc.chart
        self.loc_acc_table = loc_acc.table_widget

        # Calibration Score
        cal_score = CalibrationScore(self.vis_texts, self.comparison.eval_results)
        self.cal_score_md = cal_score.header_md
        self.cal_score_md_2 = cal_score.header_md_2
        self.cal_score_collapse_tip = cal_score.collapse_tip
        self.cal_score_table = cal_score.table
        self.cal_score_reliability_diagram_md = cal_score.reliability_diagram_md
        self.cal_score_reliability_chart = cal_score.reliability_chart
        self.cal_score_collapse_ece = cal_score.collapse_ece
        self.cal_score_confidence_score_md = cal_score.confidence_score_md
        self.cal_score_confidence_chart = cal_score.confidence_chart
        self.cal_score_confidence_score_md_2 = cal_score.confidence_score_md_2
        self.cal_score_collapse_conf_score = cal_score.collapse_conf_score

        # SpeedTest
        self.speedtest_present = False
        self.speedtest_multiple_batch_sizes = False
        if not speedtest.is_empty():
            self.speedtest_present = True
            self.speedtest_md_intro = speedtest.md_intro
            self.speedtest_intro_table = speedtest.intro_table
            self.speed_inference_time_md = speedtest.inference_time_md
            self.speed_inference_time_table = speedtest.inference_time_table
            self.speed_fps_md = speedtest.fps_md
            self.speed_fps_table = speedtest.fps_table
            self.speedtest_multiple_batch_sizes = speedtest.multiple_batche_sizes()
            if self.speedtest_multiple_batch_sizes:
                self.speed_batch_inference_md = speedtest.batch_inference_md
                self.speed_chart = speedtest.chart

    def _create_layout(self):
        is_anchors_widgets = [
            # Overview
            (0, self.header),
            (1, self.overview_md),
            (0, self.overviews),
        ]

        if self.iou_per_class_thresholds_md is not None:
            is_anchors_widgets.append((0, self.iou_per_class_thresholds_md))

        is_anchors_widgets += [
            # Key Metrics
            (1, self.key_metrics_md),
            (0, self.key_metrics_table),
            (0, self.overview_chart),
            # Explore Predictions
            (1, self.explore_predictions_md),
            (0, self.explore_predictions_gallery),
            # Outcome Counts
            (1, self.outcome_counts_md),
            (0, self.outcome_counts_main),
            (0, self.outcome_counts_diff_md),
            (0, self.outcome_counts_comparison),
            # Precision-Recall Curve
            (1, self.pr_curve_md),
            (0, self.pr_curve_collapsed_widgets),
            (0, self.pr_curve_table),
            (0, self.pr_curve_chart),
            # Average Precision by Class
            (1, self.avg_prec_by_class_md),
            (0, self.avg_prec_by_class_chart),
            # Precision, Recall, F1
            (1, self.precision_recall_f1_md),
            (0, self.precision_recall_f1_table),
            (0, self.clickable_label),
            (0, self.precision_recall_f1_chart),
            (0, self.precision_per_class_title_md),
            (0, self.precision_per_class_chart),
            (0, self.recall_per_class_title_md),
            (0, self.recall_per_class_chart),
            (0, self.f1_per_class_title_md),
            (0, self.f1_per_class_chart),
            # Classification Accuracy # TODO
            # Localization Accuracy (IoU)
            (1, self.loc_acc_header_md),
            (0, self.loc_acc_table),
            (0, self.loc_acc_iou_distribution_md),
            (0, self.loc_acc_chart),
            # Calibration Score
            (1, self.cal_score_md),
            (0, self.cal_score_md_2),
            (0, self.cal_score_collapse_tip),
            (0, self.cal_score_table),
            (1, self.cal_score_reliability_diagram_md),
            (0, self.cal_score_reliability_chart),
            (0, self.cal_score_collapse_ece),
            (1, self.cal_score_confidence_score_md),
            (0, self.cal_score_confidence_chart),
            (0, self.cal_score_confidence_score_md_2),
            (0, self.cal_score_collapse_conf_score),
        ]
        if self.speedtest_present:
            is_anchors_widgets.extend(
                [
                    # SpeedTest
                    (1, self.speedtest_md_intro),
                    (0, self.speedtest_intro_table),
                    (0, self.speed_inference_time_md),
                    (0, self.speed_inference_time_table),
                    (0, self.speed_fps_md),
                    (0, self.speed_fps_table),
                ]
            )
            if self.speedtest_multiple_batch_sizes:
                is_anchors_widgets.append((0, self.speed_batch_inference_md))
                is_anchors_widgets.append((0, self.speed_chart))
        anchors = []
        for is_anchor, widget in is_anchors_widgets:
            if is_anchor:
                anchors.append(widget.id)

        sidebar = SidebarWidget(widgets=[i[1] for i in is_anchors_widgets], anchors=anchors)
        layout = ContainerWidget(
            widgets=[sidebar, self.explore_modal_table, self.explore_predictions_modal_gallery],
            name="main_container",
        )
        return layout

    def _create_key_metrics(self) -> MarkdownWidget:
        key_metrics_text = self.vis_texts.markdown_key_metrics.format(
            self.vis_texts.definitions.average_precision,
            self.vis_texts.definitions.confidence_threshold,
            self.vis_texts.definitions.confidence_score,
        )
        return MarkdownWidget("markdown_key_metrics", "Key Metrics", text=key_metrics_text)

    def _create_outcome_counts_md(self) -> MarkdownWidget:
        outcome_counts_text = self.vis_texts.markdown_outcome_counts.format(
            self.vis_texts.definitions.true_positives,
            self.vis_texts.definitions.false_positives,
            self.vis_texts.definitions.false_negatives,
        )
        return MarkdownWidget("markdown_outcome_counts", "Outcome Counts", text=outcome_counts_text)

    def _create_outcome_counts_diff_md(self) -> MarkdownWidget:
        outcome_counts_text = self.vis_texts.markdown_outcome_counts_diff
        return MarkdownWidget(
            "markdown_outcome_counts_diff", "Outcome Counts Differences", text=outcome_counts_text
        )



from supervisely.cli.release.release import (
    slug_is_valid,
    get_module_root,
    get_module_path,
    get_remote_url,
    find_tag_in_repo,
    get_appKey,
    get_app_from_instance,
    upload_archive,
    archive_application,
    release,
    delete_tag,
    push_tag,
)
from supervisely.cli.release.run import run, cli_run



from typing import Iterable, List, Optional, Union

import numpy as np

from supervisely.annotation.annotation import Annotation
from supervisely.geometry.rectangle import Rectangle
from supervisely.nn.model.model_api import Prediction
from supervisely.video_annotation.frame import Frame
from supervisely.video_annotation.frame_collection import FrameCollection
from supervisely.video_annotation.video_annotation import VideoAnnotation
from supervisely.video_annotation.video_figure import VideoFigure
from supervisely.video_annotation.video_object import VideoObject
from supervisely.video_annotation.video_object_collection import VideoObjectCollection


def _none_generator():
    while True:
        yield None


def apply_boxmot(
    tracker,
    predictions: Union[List[Prediction], List[Annotation]],
    class_names: List[str],
    frames: Optional[Iterable[np.ndarray]] = None,
) -> VideoAnnotation:
    if frames is None:
        frames = _none_generator()
    results = []
    annotations = []
    frames_count = 0
    for prediction, frame in zip(predictions, frames):
        frames_count += 1
        if isinstance(prediction, Prediction):
            annotation = prediction.annotation
            if frame is None:
                frame = prediction.load_image()
        else:
            annotation = prediction
        frame_shape = frame.shape[:2]
        annotations.append(annotation)
        detections = to_boxes(annotation, class_names)  # N x (x, y, x, y, conf, label)
        tracks = tracker.update(
            detections, frame
        )  # M x (x, y, x, y, track_id, conf, label, det_id)
        results.append(tracks)
    return create_video_annotation(annotations, results, class_names, frame_shape, frames_count)


def to_boxes(ann: Annotation, class_names: List[str]) -> np.ndarray:
    """
    Convert annotation to detections array in boxmot format.
    :param ann: Supervisely Annotation object
    :type ann: Annotation
    :param class_names: model class names
    :type class_names: List[str]
    :return: detections array N x (x, y, x, y, conf, label)
    :rtype: np.ndarray
    """
    # convert ann to N x (x, y, x, y, conf, cls) np.array
    cls2label = {class_name: i for i, class_name in enumerate(class_names)}
    detections = []
    for label in ann.labels:
        cat = cls2label[label.obj_class.name]
        bbox = label.geometry.to_bbox()
        conf = label.tags.get("confidence").value
        detections.append([bbox.left, bbox.top, bbox.right, bbox.bottom, conf, cat])
    detections = np.array(detections)
    return detections


def create_video_annotation(
    annotations: List[Annotation],
    tracking_results: list,
    class_names: List[str],
    frame_shape: tuple,
    frames_count: int,
) -> VideoAnnotation:
    img_h, img_w = frame_shape
    video_objects = {}  # track_id -> VideoObject
    frames = []
    cat2obj = {}
    name2cat = {class_name: i for i, class_name in enumerate(class_names)}
    obj_classes = {}
    for annotation in annotations:
        for label in annotation.labels:
            obj_classes.setdefault(label.obj_class.name, label.obj_class)
    for obj_name, cat in name2cat.items():
        obj_class = obj_classes.get(obj_name)
        if obj_class is None:
            raise ValueError(f"Object class {obj_name} not found in annotations.")
        cat2obj[cat] = obj_class
    for i, tracks in enumerate(tracking_results):
        frame_figures = []
        for track in tracks:
            # crop bbox to image size
            dims = np.array([img_w, img_h, img_w, img_h]) - 1
            track[:4] = np.clip(track[:4], 0, dims)
            x1, y1, x2, y2, track_id, conf, cat = track[:7]
            cat = int(cat)
            track_id = int(track_id)
            rect = Rectangle(y1, x1, y2, x2)
            video_object = video_objects.setdefault(track_id, VideoObject(cat2obj[cat]))
            frame_figures.append(VideoFigure(video_object, rect, i))
        frames.append(Frame(i, frame_figures))

    objects = list(video_objects.values())
    video_annotation = VideoAnnotation(
        img_size=frame_shape,
        frames_count=frames_count,
        objects=VideoObjectCollection(objects),
        frames=FrameCollection(frames),
    )
    return video_annotation



docs_url = (
    "https://docs.supervisely.com/neural-networks/model-evaluation-benchmark/semantic-segmentation"
)

# <i class="zmdi zmdi-check-circle" style="color: #13ce66; margin-right: 5px"></i>
clickable_label = """
> <span style="color: #5a6772">
>     Click on the chart to explore corresponding images.
> </span>
"""

markdown_header = """
<h1>{}</h1>

<div class="model-info-block">
    <div>Created by <b>{}</b></div>
    <div><i class="zmdi zmdi-calendar-alt"></i><span>{}</span></div>
</div>
"""

markdown_common_overview = """
- **Models**: {}
- **Evaluation Dataset**: <a href="/projects/{}/datasets" target="_blank">{}</a>
- **Task type**: {}
"""

markdown_overview_info = """
<h3>{}</h3>
- **Model**: {}
- **Checkpoint**: {}
- **Architecture**: {}
- **Runtime**: {}
- **Checkpoint file**: <a class="checkpoint-url" href="{}" target="_blank">{}</a>
- **Evaluation Report**: <a href="{}" target="_blank">View Report</a>

"""

markdown_key_metrics = """## Key Metrics

We provide a comprehensive analysis of models' performance using a set of metrics, including both basic (precision, recall, F1-score, IoU, etc.) and advanced (boundary IoU, error over union decomposition, etc.) metrics.

- **Pixel accuracy**: reflects the percent of image pixels which were correctly classified.
- **Precision**: reflects the number of correctly predicted positive segmentations divided by the total number of predicted positive segmentations.
- **Recall**: reflects the number of correctly predicted positive segmentations divided by the number of all samples that should have been segmented as positive.
- **F1-score**: reflects the tradeoff between precision and recall. It is equivalent to the Dice coefficient and calculated as a harmonic mean of precision and recall.
- **Intersection over union (IoU, also known as the Jaccard index)**: measures the overlap between ground truth mask and predicted mask. It is calculated as the ratio of the intersection of the two masks areas to their combined areas.
- **Boundary intersection over union**: a segmentation consistency measure that first computes the sets of ground truth  and predicted masks pixels that are located within the distance d from each contour and then computes intersection over union of these two sets. Pixel distance parameter d (pixel width of the boundary region) controls the sensitivity of the metric, it is usually set as 2% of the image diagonal for normal resolution images and 0.5% of the image diagonal for high resolution images.
- **Error over union and its components (boundary, extent, segment)**: a metric opposite to intersection over union and can be interpreted as what the model lacked in order to show the perfect performance with IoU = 1. It reflects the ratio of incorrectly segmented pixels of ground truth and predicted masks to their combined areas. It is usually decomposed into boundary, extent and segment errors over union in order to get exhaustive information about the model's strengths and weaknesses.
- **Renormalized error over union**: postprocessed variant of error over union which takes into consideration cause and effect relationships between different types of segmentation errors.
"""

markdown_explorer = """## Explore Predictions
This section contains visual comparison of predictions made by different models and ground truth annotations. Sometimes a simple visualization can be more informative than any performance metric.

> Click on the image to view the **Original Image** with **Ground Truth** and **Predictions** annotations side-by-side. 
"""

markdown_explore_difference = """## Explore Predictions

In this section, you can explore predictions made by different models side-by-side. This helps you to understand the differences in predictions made by each model, and to identify which model performs better in different scenarios.


> Click on the image to view the **Ground Truth**, and **Prediction** annotations side-by-side.
"""


### Difference in Predictions

# markdown_explore_same_errors = """
# ### Same Errors

# This section helps you to identify samples where all models made the same errors. It is useful for understanding the limitations of the models and the common challenges they face.

# > Click on the image to view the **Ground Truth**, and **Prediction** annotations side-by-side.
# """


# """
markdown_iou = """## Intersection & Error Over Union

Pie charts below demonstrate performance metrics of each model in terms of Intersection over Union (IoU) and Error over Union (EoU). It is done with the help of Error over Union (EoU) decomposition into boundary, extent, and segment errors over union. These charts help to draw conclusions on the model's strongest and weakest sides.
"""

markdown_renormalized_error_ou = """## Renormalized Error over Union

Charts below are dedicated to the decomposition of the post-processed variant of Error over Union, which takes into consideration cause and effect relationships between different types of segmentation errors. Error over Union decomposition has its own pitfalls. It is important to understand that models which tend to produce segment errors (when entire segments are mispredicted and there is no intersection between ground truth and predicted mask) will face fewer occasions to produce boundary and extent errors - as a result, boundary and extent error over union values will be underestimated.

In terms of localization, segment error is more fundamental than extent, while extent error is more fundamental than boundary. In order to overcome this problem, renormalized error over union proposes a slightly different calculation method - by removing more fundamental errors from the denominator - read more in our <a href="{}" target="_blank">technical report</a>
""".format(
    docs_url
)

markdown_eou_per_class = """## Classwise Segmentation Error Analysis

This section contains information about classwise segmentation error decomposition. For each model, each column of the chart represents a certain class from the training dataset, demonstrating model performance in terms of segmenting this specific class on images and what model lacked in order to show the perfect performance.
"""

markdown_frequently_confused_empty = """### Frequently Confused Classes

No frequently confused class pairs found
"""

markdown_frequently_confused = """## Frequently Confused Classes

The bar chart below reveals pairs of classes which were most frequently confused for each model. Each column of the chart demonstrates the probability of confusion of a given pair of classes. It is necessary to remember that this probability is not symmetric: the probability of confusing class A with class B is not equal to the probability of confusing class B with class A.
"""

empty = """### {}

> {}
"""

markdown_speedtest_intro = """## Inference Speed

This is a speed test benchmark for compared models. Models were tested with the following configurations:
"""

markdown_speedtest_overview_ms = """### Latency (Inference Time)
The table below shows the speed test results. For each test, the time taken to process one batch of images is shown. Results are averaged across **{}** iterations.
"""

markdown_speedtest_overview_fps = """### Frames per Second (FPS)
The table below shows the speed test results. For each test, the number of frames processed per second is shown. Results are averaged across **{}** iterations.
"""

markdown_batch_inference = """
This chart shows how the model's speed changes with different batch sizes . As the batch size increases, you can observe an increase in FPS (images per second).
"""



from __future__ import annotations

from supervisely.nn.benchmark.object_detection.base_vis_metric import DetectionVisMetric
from supervisely.nn.benchmark.visualization.widgets import (
    ChartWidget,
    CollapseWidget,
    MarkdownWidget,
    NotificationWidget,
)


class ConfidenceScore(DetectionVisMetric):
    MARKDOWN_CONFIDENCE_SCORE = "confidence_score"
    NOTIFICATION = "confidence_score"
    MARKDOWN_CONFIDENCE_SCORE_2 = "confidence_score_2"
    MARKDOWN_CONFIDENCE_SCORE_3 = "calibration_score_3"
    COLLAPSE_TITLE = "confidence_score_collapse"
    CHART = "confidence_score"

    @property
    def md_confidence_score(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_confidence_score_1
        text = text.format(self.vis_texts.definitions.confidence_threshold)
        return MarkdownWidget(self.MARKDOWN_CONFIDENCE_SCORE, "Confidence Score Profile", text)

    @property
    def notification(self) -> NotificationWidget:
        title, _ = self.vis_texts.notification_f1.values()
        return NotificationWidget(
            self.NOTIFICATION,
            title.format(self.eval_result.mp.f1_optimal_conf.round(4)),
        )

    @property
    def chart(self) -> ChartWidget:
        return ChartWidget(name=self.CHART, figure=self._get_figure())

    @property
    def md_confidence_score_2(self) -> MarkdownWidget:
        return MarkdownWidget(
            self.MARKDOWN_CONFIDENCE_SCORE_2,
            "",
            self.vis_texts.markdown_confidence_score_2,
        )

    @property
    def collapse_conf_score(self) -> CollapseWidget:
        md = MarkdownWidget(
            self.COLLAPSE_TITLE,
            "How to plot Confidence Profile?",
            self.vis_texts.markdown_plot_confidence_profile,
        )
        return CollapseWidget([md])

    @property
    def md_confidence_score_3(self) -> MarkdownWidget:
        return MarkdownWidget(
            self.MARKDOWN_CONFIDENCE_SCORE_3,
            "",
            self.vis_texts.markdown_calibration_score_3,
        )

    def _get_figure(self):  # -> go.Figure:
        import plotly.express as px  # pylint: disable=import-error

        color_map = {
            "Precision": "#1f77b4",
            "Recall": "orange",
        }

        fig = px.line(
            self.eval_result.dfsp_down,
            x="scores",
            y=["precision", "recall", "f1"],
            labels={"value": "Metric", "variable": "Metric", "scores": "Confidence Score"},
            width=None,
            height=500,
            color_discrete_map=color_map,
        )
        fig.update_traces(
            hovertemplate="Confidence score: %{x:.2f}<br>Metric: %{y:.2f}<extra></extra>"
        )
        fig.update_layout(yaxis=dict(range=[0, 1]), xaxis=dict(range=[0, 1], tick0=0, dtick=0.1))

        if (
            self.eval_result.mp.f1_optimal_conf is not None
            and self.eval_result.mp.best_f1 is not None
        ):
            # Add vertical line for the best threshold
            fig.add_shape(
                type="line",
                x0=self.eval_result.mp.f1_optimal_conf,
                x1=self.eval_result.mp.f1_optimal_conf,
                y0=0,
                y1=self.eval_result.mp.best_f1,
                line=dict(color="gray", width=2, dash="dash"),
            )
            fig.add_annotation(
                x=self.eval_result.mp.f1_optimal_conf,
                y=self.eval_result.mp.best_f1 + 0.04,
                text=f"F1-optimal threshold: {self.eval_result.mp.f1_optimal_conf:.2f}",
                showarrow=False,
            )
        if self.eval_result.mp.custom_conf_threshold is not None:
            # Add vertical line for the custom threshold
            fig.add_shape(
                type="line",
                x0=self.eval_result.mp.custom_conf_threshold,
                x1=self.eval_result.mp.custom_conf_threshold,
                y0=0,
                y1=self.eval_result.mp.custom_f1,
                line=dict(color="black", width=2, dash="dash"),
            )
            fig.add_annotation(
                x=self.eval_result.mp.custom_conf_threshold,
                y=self.eval_result.mp.custom_f1 + 0.04,
                text=f"Confidence threshold: {self.eval_result.mp.custom_conf_threshold:.2f}",
                showarrow=False,
            )
        fig.update_layout(
            dragmode=False,
            modebar=dict(
                remove=[
                    "zoom2d",
                    "pan2d",
                    "select2d",
                    "lasso2d",
                    "zoomIn2d",
                    "zoomOut2d",
                    "autoScale2d",
                    "resetScale2d",
                ]
            ),
        )
        return fig



import time
import uuid
from collections import OrderedDict, namedtuple
from logging import Logger
from queue import Queue
from threading import Lock, Thread
from typing import Any, Callable, Dict, Generator, List, Optional
from typing import OrderedDict as OrderedDictType

import numpy as np

from supervisely._utils import find_value_by_keys
from supervisely.api.api import Api
from supervisely.api.module_api import ApiField
from supervisely.geometry.geometry import Geometry
from supervisely.geometry.graph import GraphNodes
from supervisely.geometry.helpers import deserialize_geometry
from supervisely.geometry.point import Point
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.polyline import Polyline
from supervisely.geometry.rectangle import Rectangle
from supervisely.nn.inference.cache import InferenceImageCache
from supervisely.sly_logger import logger
from supervisely.video_annotation.key_id_map import KeyIdMap


class TrackerInterface:
    def __init__(
        self,
        context,
        api,
        load_all_frames=False,
        notify_in_predict=False,
        per_point_polygon_tracking=True,
        frame_loader: Callable[[Api, int, int], np.ndarray] = None,
        frames_loader: Callable[[Api, int, List[int]], List[np.ndarray]] = None,
        should_notify: bool = True,
    ):
        self.api: Api = api
        self.logger: Logger = api.logger
        self.frame_index = context["frameIndex"]
        self.frames_count = context["frames"]

        self.track_id = context["trackId"]
        self.video_id = context["videoId"]
        self.frame_width = context.get("frameWidth", None)
        self.frame_height = context.get("frameHeight", None)
        self._video_info = None
        self.object_ids = list(context["objectIds"])
        self.figure_ids = list(context["figureIds"])
        self.direction = context["direction"]
        self.should_notify = should_notify

        # all geometries
        self.stop = len(self.figure_ids) * self.frames_count
        self.global_pos = 0
        self.global_stop_indicatior = False

        self.geometries: OrderedDictType[int, Geometry] = OrderedDict()
        self.frames_indexes: List[int] = []
        self._cur_frames_indexes: List[int] = []
        self._frames: Optional[np.ndarray] = None
        self.load_all_frames = load_all_frames
        self.per_point_polygon_tracking = per_point_polygon_tracking

        # increase self.stop by num of frames will be loaded
        self._add_frames_indexes()

        # increase self.stop by num of points
        self._add_geometries()

        self._hot_cache: Dict[int, np.ndarray] = {}
        self._local_cache_loader = frame_loader
        self._local_cache_frames_loader = frames_loader

        if self.load_all_frames:
            if notify_in_predict:
                self.stop += self.frames_count + 1
            self._load_frames_to_hot_cache()
            self._load_frames()

    @property
    def video_info(self):
        if self._video_info is None:
            self._video_info = self.api.video.get_info_by_id(self.video_id)
        return self._video_info

    def add_object_geometries(self, geometries: List[Geometry], object_id: int, start_fig: int):
        for frame, geometry in zip(self._cur_frames_indexes[1:], geometries):
            if self.global_stop_indicatior:
                self._notify(True, task="stop tracking")
                break
            self.add_object_geometry_on_frame(geometry, object_id, frame)

        self.geometries[start_fig] = geometries[-1]

    def frames_loader_generator(self, batch_size=4) -> Generator[None, None, None]:
        if self.load_all_frames:
            self._cur_frames_indexes = self.frames_indexes
            yield
            return

        self._load_frames_to_hot_cache(
            self.frames_indexes[: min(batch_size + 1, len(self.frames_indexes))]
        )
        ind = self.frames_indexes[0]
        frame = self._load_frame(ind)
        for next_ind_pos, next_ind in enumerate(self.frames_indexes[1:]):
            if next_ind not in self._hot_cache:
                self._load_frames_to_hot_cache(
                    self.frames_indexes[
                        next_ind_pos
                        + 1 : min(next_ind_pos + 1 + batch_size, len(self.frames_indexes))
                    ]
                )
            next_frame = self._load_frame(next_ind)
            self._frames = np.array([frame, next_frame])
            self.frames_count = 1
            self._cur_frames_indexes = [ind, next_ind]
            yield
            frame = next_frame
            ind = next_ind

            if self.global_stop_indicatior:
                self.clear_cache()
                return

    def _crop_geometry(self, geometry: Geometry):
        h, w = self.frame_height, self.frame_width
        if h is None or w is None:
            h = self.video_info.frame_height
            w = self.video_info.frame_width
        rect = Rectangle.from_size((h, w))
        cropped = geometry.crop(rect)
        if len(cropped) == 0:
            return None
        return cropped[0]

    def add_object_geometries_on_frames(
        self,
        geometries: List[Geometry],
        object_ids: List[int],
        frame_indexes: List[int],
        notify: bool = True,
    ):
        def _split(geometries: List[Geometry], object_ids: List[int], frame_indexes: List[int]):
            result = {}
            for geometry, object_id, frame_index in zip(geometries, object_ids, frame_indexes):
                result.setdefault(object_id, []).append((geometry, frame_index))
            return result

        geometries_by_object = _split(geometries, object_ids, frame_indexes)

        for object_id, geometries_frame_indexes in geometries_by_object.items():
            for i, (geometry, frame_index) in enumerate(geometries_frame_indexes):
                geometries_frame_indexes[i] = (self._crop_geometry(geometry), frame_index)
            geometries_frame_indexes = [
                (geometry, frame_index)
                for geometry, frame_index in geometries_frame_indexes
                if geometry is not None
            ]
            figures_json = [
                {
                    ApiField.OBJECT_ID: object_id,
                    ApiField.GEOMETRY_TYPE: geometry.geometry_name(),
                    ApiField.GEOMETRY: geometry.to_json(),
                    ApiField.META: {ApiField.FRAME: frame_index},
                    ApiField.TRACK_ID: self.track_id,
                }
                for geometry, frame_index in geometries_frame_indexes
            ]
            figures_keys = [uuid.uuid4() for _ in figures_json]
            key_id_map = KeyIdMap()
            self.api.video.figure._append_bulk(
                entity_id=self.video_id,
                figures_json=figures_json,
                figures_keys=figures_keys,
                key_id_map=key_id_map,
            )
            self.logger.debug(f"Added {len(figures_json)} geometries to object #{object_id}")
            if notify:
                self._notify(task="add geometry on frame", pos_increment=len(figures_json))

    def add_object_geometry_on_frame(
        self,
        geometry: Geometry,
        object_id: int,
        frame_ind: int,
        notify: bool = True,
    ):
        self.api.video.figure.create(
            self.video_id,
            object_id,
            frame_ind,
            geometry.to_json(),
            geometry.geometry_name(),
            self.track_id,
        )
        self.logger.debug(f"Added {geometry.geometry_name()} to frame #{frame_ind}")
        if notify:
            self._notify(task="add geometry on frame")

    def clear_cache(self):
        self._hot_cache.clear()

    def _add_geometries(self):
        self.logger.info("Adding geometries.")
        points = 0
        for figure_id in self.figure_ids:
            figure = self.api.video.figure.get_info_by_id(figure_id)
            geometry = deserialize_geometry(figure.geometry_type, figure.geometry)
            self.geometries[figure_id] = geometry

            self.api.logger.debug(f"Added {figure.geometry_type} #{figure_id}")

            # per point track notification
            if isinstance(geometry, Point):
                points += 1
            elif isinstance(geometry, Polygon):
                points += len(geometry.exterior) + len(geometry.interior)
            elif isinstance(geometry, GraphNodes):
                points += len(geometry.nodes.items())
            elif isinstance(geometry, Polyline):
                points += len(geometry.exterior)

        if self.per_point_polygon_tracking:
            if not self.load_all_frames:
                self.stop += points * self.frames_count
            else:
                self.stop += points

        self.logger.info("Geometries added.")
        # TODO: other geometries

    def _add_frames_indexes(self):
        total_frames = self.api.video.get_info_by_id(self.video_id).frames_count
        cur_index = self.frame_index

        while 0 <= cur_index < total_frames and len(self.frames_indexes) < self.frames_count + 1:
            self.frames_indexes.append(cur_index)
            cur_index += 1 if self.direction == "forward" else -1

        if self.load_all_frames:
            self.stop += len(self.frames_indexes)

    def _load_frames_to_hot_cache(self, frames_indexes: List[int] = None):
        if self._local_cache_frames_loader is not None:
            if frames_indexes is None:
                frames_indexes = self.frames_indexes
            frames_to_load = []
            for frame_index in frames_indexes:
                if frame_index not in self._hot_cache:
                    frames_to_load.append(frame_index)
            if len(frames_to_load) == 0:
                return
            self.logger.info(f"Loading {frames_to_load} frames to hot cache.")
            loaded_rgbs = self._local_cache_frames_loader(self.api, self.video_id, frames_to_load)
            for rgb, loaded_frame_index in zip(loaded_rgbs, frames_to_load):
                self._hot_cache[loaded_frame_index] = rgb
            self.logger.info(f"{frames_to_load} frames loaded to hot cache.")

    def _load_frame(self, frame_index):
        if frame_index in self._hot_cache:
            return self._hot_cache[frame_index]
        if self._local_cache_loader is None:
            self._hot_cache[frame_index] = self.api.video.frame.download_np(
                self.video_id, frame_index
            )
            return self._hot_cache[frame_index]
        else:
            return self._local_cache_loader(self.api, self.video_id, frame_index)

    def _load_frames(self):
        rgbs = []
        self.logger.info(f"Loading {len(self.frames_indexes)} frames.")

        for frame_index in self.frames_indexes:
            img_rgb = self._load_frame(frame_index)
            rgbs.append(img_rgb)
            self._notify(task="load frame")
        self._frames = rgbs
        self.logger.info("Frames loaded.")

    def _notify(
        self,
        stop: bool = False,
        fstart: Optional[int] = None,
        fend: Optional[int] = None,
        task: str = "not defined",
        pos_increment: int = 1,
    ):
        if not self.should_notify:
            return

        self.global_pos += pos_increment

        if stop:
            pos = self.stop
        else:
            pos = self.global_pos

        fstart = min(self.frames_indexes) if fstart is None else fstart
        fend = max(self.frames_indexes) if fend is None else fend

        self.logger.debug(f"Task: {task}")
        self.logger.debug(f"Notification status: {pos}/{self.stop}")

        self.global_stop_indicatior = self.api.video.notify_progress(
            self.track_id,
            self.video_id,
            fstart,
            fend,
            pos,
            self.stop,
        )

        self.logger.debug(f"Notification status: stop={self.global_stop_indicatior}")

        if self.global_stop_indicatior and self.global_pos < self.stop:
            self.logger.info("Task stoped by user.")

    @property
    def frames(self) -> np.ndarray:
        return self._frames

    @property
    def frames_with_notification(self) -> np.ndarray:
        """Use this in prediction."""
        self._notify(task="get frames")
        return self._frames


class ThreadSafeStopIndicator:
    def __init__(self):
        self._stopped = False
        self._reason = None
        self._lock = Lock()

    def stop(self, reason: Any = None):
        if self.is_stopped():
            return
        with self._lock:
            self._stopped = True
            self._reason = reason

    def is_stopped(self):
        with self._lock:
            return self._stopped

    def get_reason(self):
        with self._lock:
            return self._reason


FrameImage = namedtuple("FrameImage", ["frame_index", "image"])


class TrackerInterfaceV2:
    UPLOAD_SLEEP_TIME = 0.001  # 1ms
    NOTIFY_SLEEP_TIME = 1  # 1s

    def __init__(
        self,
        api: Api,
        context: Dict,
        cache: InferenceImageCache,
    ):
        self.api = api
        self.context = context
        self.video_id = find_value_by_keys(context, ["videoId", "video_id"])
        self.frame_index = find_value_by_keys(context, ["frameIndex", "frame_index"])
        self.frames_count = find_value_by_keys(context, ["frames", "framesCount", "frames_count"])
        self.track_id = context.get("trackId", "auto")
        self.direction = context.get("direction", "forward")
        self.session_id = find_value_by_keys(context, ["sessionId", "session_id"], None)
        self.figures = context.get("figures", None)
        self.direct_progress = context.get("useDirectProgressMessages", False)
        self.direction_n = 1 if self.direction == "forward" else -1
        self.stop_indicator = ThreadSafeStopIndicator()
        self.cache = cache
        self.frame_indexes = list(
            range(
                self.frame_index,
                self.frame_index + self.frames_count * self.direction_n + self.direction_n,
                self.direction_n,
            )
        )

        self.log_extra = {
            "video_id": self.video_id,
            "track_id": self.track_id,
            "session_id": self.session_id,
        }

        # start caching task
        self.run_cache_frames_task()

        self.upload_sleep_time = self.UPLOAD_SLEEP_TIME
        self.notify_sleep_time = self.NOTIFY_SLEEP_TIME
        self.upload_queue = Queue()
        self.notify_queue = Queue()
        self.upload_thread = None
        self.notify_thread = None
        self._upload_f = None
        self._notify_f = None

    def __call__(self, upload_f, notify_f):
        self.upload_f = upload_f
        self.notify_f = notify_f
        return self

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.wait_stop(exception=exc_value)
        return False

    @property
    def upload_f(self):
        return self._upload_f

    @upload_f.setter
    def upload_f(self, upload_f):
        self._upload_f = upload_f
        self.upload_thread = Thread(
            target=self._upload_loop,
            args=[
                self.upload_queue,
                self.notify_queue,
                self.stop_indicator,
                self._upload_f,
                self.upload_sleep_time,
                self._upload_exception_handler,
            ],
            daemon=True,
        )

    @property
    def notify_f(self):
        return self._notify_f

    @notify_f.setter
    def notify_f(self, notify_f):
        self._notify_f = notify_f
        self.notify_thread = Thread(
            target=self._nofify_loop,
            args=[
                self.notify_queue,
                self.stop_indicator,
                self._notify_f,
                self.notify_sleep_time,
                self._notify_exception_handler,
            ],
            daemon=True,
        )

    def start(self):
        if self.upload_thread is not None:
            self.upload_thread.start()
        if self.notify_thread is not None:
            self.notify_thread.start()

    def stop(self, exception: Exception = None):
        self.stop_indicator.stop(exception)

    def join(self, timeout: Optional[float] = 5 * 60):
        if self.upload_thread is not None and self.upload_thread.is_alive():
            self.upload_thread.join(timeout=timeout)
        if self.notify_thread is not None and self.notify_thread.is_alive():
            self.notify_thread.join(timeout=timeout)

    def wait_stop(self, exception: Exception = None, timeout: Optional[float] = 5 * 60):
        self.stop(exception)
        self.join(timeout=timeout)

    def is_stopped(self):
        return self.stop_indicator.is_stopped()

    def stop_reason(self):
        return self.stop_indicator.get_reason()

    @classmethod
    def _upload_loop(
        cls,
        q: Queue,
        notify_q: Queue,
        stop_indicator: ThreadSafeStopIndicator,
        upload_f: callable = None,
        upload_sleep_time: float = None,
        exception_handler: callable = None,
    ):
        logger.debug("Upload loop started")
        if upload_f is None:
            logger.warning("Upload function is not provided. Exiting upload loop.")
            return
        upload_sleep_time = upload_sleep_time or cls.UPLOAD_SLEEP_TIME
        try:
            while True:
                items = []
                while not q.empty():
                    items.append(q.get_nowait())
                if len(items) > 0:
                    upload_f(items)
                    for item in items:
                        notify_q.put(item)
                elif stop_indicator.is_stopped():
                    logger.debug("stop event is set. returning from upload loop")
                    return
                time.sleep(upload_sleep_time)
        except Exception as e:
            if exception_handler is not None:
                e = exception_handler(e)
                if not isinstance(e, Exception):
                    return
                raise e
            logger.error("Error in upload loop: %s", str(e), exc_info=True)
            raise

    @classmethod
    def _nofify_loop(
        cls,
        q: Queue,
        stop_indicator: ThreadSafeStopIndicator,
        notify_f: callable = None,
        notify_sleep_time: float = None,
        exception_handler: callable = None,
    ):
        logger.debug("Notify loop started")
        if notify_f is None:
            logger.warning("Notify function is not provided. Exiting notify loop.")
            return
        notify_sleep_time = notify_sleep_time or cls.NOTIFY_SLEEP_TIME
        try:
            while True:
                items = []
                while not q.empty():
                    items.append(q.get_nowait())
                if len(items) > 0:
                    notify_f(items)
                elif stop_indicator.is_stopped():
                    logger.debug(f"stop event is set. returning from notify loop")
                    return
                time.sleep(notify_sleep_time)
        except Exception as e:
            if exception_handler is not None:
                e = exception_handler(e)
                if not isinstance(e, Exception):
                    return
                raise e
            logger.error("Error in notify loop: %s", str(e), exc_info=True)
            raise

    def run_cache_frames_task(self):
        if self.cache.is_persistent:
            frame_ranges = None
        else:
            frame_ranges = [
                self.frame_index,
                self.frame_index + self.frames_count * self.direction_n,
            ]
            if self.direction_n == -1:
                frame_ranges = frame_ranges[::-1]
            frame_ranges = [frame_ranges]
        self.cache.run_cache_task_manually(
            self.api,
            frame_ranges,
            video_id=self.video_id,
        )

    def frames_loader_generator(
        self, batch_size=2, step=1
    ) -> Generator[List[FrameImage], None, None]:
        step = step * self.direction_n
        batch = []
        t = time.monotonic()
        for frame_i, frame in zip(
            self.frame_indexes,
            self.cache.frames_loader(
                self.api, video_id=self.video_id, frame_indexes=self.frame_indexes
            ),
        ):
            batch.append(FrameImage(frame_i, frame))
            if len(batch) == batch_size:
                _get_indexes_str = lambda b: ", ".join(map(lambda x: str(x.frame_index), b))
                if len(batch) > 16:
                    batch_indexes_str = (
                        _get_indexes_str(batch[:8]) + " ... " + _get_indexes_str(batch[-8:])
                    )
                else:
                    batch_indexes_str = _get_indexes_str(batch)

                logger.debug(
                    f"Frames [{batch_indexes_str}] loaded. Time: {time.monotonic() - t}",
                    extra=self.log_extra,
                )
                yield batch
                t = time.monotonic()
                batch = batch[step:]

    def load_all_frames(self):
        return next(self.frames_loader_generator(batch_size=len(self.frame_indexes)))

    def add_prediction(self, item: Any):
        self.upload_queue.put(item)

    def notify_progress(
        self,
        progress_current: int,
        progress_total: int,
        frame_range: List[int] = None,
    ):
        logger.debug(
            f"Notify progress: {progress_current}/{progress_total} on frames {frame_range}",
            extra=self.log_extra,
        )
        if self.direct_progress:
            self.api.vid_ann_tool.set_direct_tracking_progress(
                self.session_id,
                self.video_id,
                self.track_id,
                frame_range=frame_range,
                progress_current=progress_current,
                progress_total=progress_total,
            )
        else:
            stopped = self.api.video.notify_progress(
                self.track_id,
                self.video_id,
                frame_range[0],
                frame_range[1],
                progress_current,
                progress_total,
            )
            if stopped and progress_current < progress_total:
                logger.info("Task stopped by user.", extra=self.log_extra)
                self.stop()

    def notify_error(self, exception: Exception):
        logger.debug(f"Notify error: {str(exception)}", extra=self.log_extra)
        error = type(exception).__name__
        message = str(exception)
        if self.direct_progress:
            self.api.vid_ann_tool.set_direct_tracking_error(
                self.session_id,
                self.video_id,
                self.track_id,
                f"{type(exception).__name__}: {str(exception)}",
            )
        else:
            self.api.video.notify_tracking_error(self.track_id, error, message)

    def notify_warning(self, message: str):
        logger.debug(f"Notify warning: {message}", extra=self.log_extra)
        if self.direct_progress:
            self.api.vid_ann_tool.set_direct_tracking_warning(
                self.session_id, self.video_id, self.track_id, message
            )
        else:
            self.api.video.notify_tracking_warning(self.track_id, self.video_id, message)

    def _upload_exception_handler(self, exception: Exception):
        logger.error(
            "Error in upload loop: %s", str(exception), exc_info=True, extra=self.log_extra
        )
        self.stop_indicator.stop(exception)
        raise exception

    def _notify_exception_handler(self, exception: Exception):
        logger.error(
            "Error in notify loop: %s", str(exception), exc_info=True, extra=self.log_extra
        )
        self.stop_indicator.stop(exception)
        raise exception



# coding: utf-8
"""create or manipulate already existing Entities Collection in Supervisely"""

# docs
from __future__ import annotations

from typing import Dict, List, NamedTuple, Optional

from supervisely.api.module_api import (
    ApiField,
    ModuleApi,
    RemoveableModuleApi,
    UpdateableModule,
)
from supervisely.sly_logger import logger


class EntitiesCollectionInfo(NamedTuple):
    id: int
    name: str
    team_id: int
    project_id: int
    description: str
    created_at: str
    updated_at: str


class EntitiesCollectionApi(UpdateableModule, RemoveableModuleApi):
    """
    API for working with Entities Collection. :class:`EntitiesCollectionApi<EntitiesCollectionApi>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervise.ly", token="4r47N...xaTatb")

        collection = api.entities_collection.get_list(9) # api usage example
    """

    @staticmethod
    def info_sequence():
        """
        NamedTuple EntitiesCollectionInfo information about Entities Collection.

        :Example:

         .. code-block:: python

             EntitiesCollectionInfo(
                id=2,
                name='Enitites Collections #1',
                team_id=4,
                project_id=58,
                description='',
                created_at='2020-04-08T15:10:12.618Z',
                updated_at='2020-04-08T15:10:19.833Z',
            )
        """
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.TEAM_ID,
            ApiField.PROJECT_ID,
            ApiField.DESCRIPTION,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
        ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **EntitiesCollectionInfo**.
        """
        return "EntitiesCollectionInfo"

    def __init__(self, api):
        ModuleApi.__init__(self, api)

    def create(
        self, project_id: int, name: str, description: Optional[str] = None
    ) -> EntitiesCollectionInfo:
        """
        Creates Entities Collections.

        :param project_id: Project ID in Supervisely.
        :type project_id: int
        :param name: Entities Collection name in Supervisely.
        :type name: str
        :param description: Entities Collection description.
        :type description: str
        :return: Information about new Entities Collection
        :rtype: :class:`EntitiesCollectionInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            user_name = 'Collection #1'
            project_id = 602
            new_collection = api.entities_collection.create(name, project_id)
            print(new_collection)
        """

        data = {ApiField.NAME: name, ApiField.PROJECT_ID: project_id}
        if description is not None:
            data[ApiField.DESCRIPTION] = description
        response = self._api.post("entities-collections.add", data)
        return self._convert_json_info(response.json())

    def get_list(
        self,
        project_id: int,
        filters: Optional[List[Dict[str, str]]] = None,
    ) -> List[EntitiesCollectionInfo]:
        """
        Get list of information about Entities Collection for the given project.

        :param project_id: Project ID in Supervisely.
        :type project_id: int, optional
        :return: List of information about Entities Collections.
        :rtype: :class:`List[EntitiesCollectionInfo]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            collections = api.entities_collection.get_list(4)
        """
        if filters is None:
            filters = []
        return self.get_list_all_pages(
            "entities-collections.list",
            {ApiField.PROJECT_ID: project_id, ApiField.FILTER: filters},
        )

    def get_info_by_id(self, id: int) -> EntitiesCollectionInfo:
        """
        Get information about Entities Collection with given ID.

        :param id: Entities Collection ID in Supervisely.
        :type id: int
        :return: Information about Entities Collection.
        :rtype: :class:`EntitiesCollectionInfo`
        :Usage example:

            .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            collection = api.entities_collection.get_info_by_id(2)
            print(collection)
            # Output:
            # {
            #     "id": 1,
            #     "teamId": 1,
            #     "projectId": 1,
            #     "name": "ds",
            #     "description": "",
            #     "createdAt": "2018-08-21T14:25:56.140Z",
            #     "updatedAt": "2018-08-21T14:25:56.140Z"
            # }
        """
        return self._get_info_by_id(id, "entities-collections.info")

    def _get_update_method(self):
        """ """
        return "entities-collections.editInfo"

    def _remove_api_method_name(self):
        """ """
        return "entities-collections.remove"

    def _get_update_method(self):
        """ """
        return "entities-collections.editInfo"

    def add_items(self, id: int, items: List[int]) -> List[Dict[str, int]]:
        """
        Add items to Entities Collection.

        :param id: Entities Collection ID in Supervisely.
        :type id: int
        :param items: List of item IDs in Supervisely.
        :type items: List[int]
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            collection_id = 2
            item_ids = [525, 526]
            new_items = api.entities_collection.add_items(collection_id, item_ids)
            print(new_items)
            # Output: [
            #   {"id": 1, "entityId": 525, 'createdAt': '2025-04-10T08:49:41.852Z'},
            #   {"id": 2, "entityId": 526, 'createdAt': '2025-04-10T08:49:41.852Z'}
            ]
        """
        data = {ApiField.COLLECTION_ID: id, ApiField.ENTITY_IDS: items}
        response = self._api.post("entities-collections.items.bulk.add", data)
        response = response.json()
        if len(response["missing"]) > 0:
            raise RuntimeError(
                f"Failed to add items to Entities Collection. IDs: {response['missing']}. "
            )
        return response["items"]

    def get_items(self, id: int, project_id: Optional[int] = None) -> List[int]:
        """
        Get items from Entities Collection.

        :param id: Entities Collection ID in Supervisely.
        :type id: int
        :param project_id: Project ID in Supervisely.
        :type project_id: int, optional
        :return: List of item IDs in Supervisely.
        :rtype: List[int]
        :raises RuntimeError: If Entities Collection with given ID not found.
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            collection_id = 2
            project_id = 4
            item_ids = api.entities_collection.get_items(collection_id, project_id)
            print(item_ids)
        """
        if project_id is None:
            info = self.get_info_by_id(id)
            if info is None:
                raise RuntimeError(f"Entities Collection with id={id} not found.")
            project_id = info.project_id

        return self._api.image.get_list(project_id=project_id, entities_collection_id=id)

    def remove_items(self, id: int, items: List[int]) -> List[Dict[str, int]]:
        """
        Remove items from Entities Collection.

        :param id: Entities Collection ID in Supervisely.
        :type id: int
        :param items: List of item IDs in Supervisely.
        :type items: List[int]
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            api = sly.Api.from_env()

            collection_id = 2
            item_ids = [525, 526, 527]
            removed_items = api.entities_collection.remove_items(collection_id, item_ids)
            # print(removed_items)
            # Output: [{"id": 1, "entityId": 525}, {"id": 2, "entityId": 526}]
        """
        data = {ApiField.COLLECTION_ID: id, ApiField.ENTITY_IDS: items}
        response = self._api.post("entities-collections.items.bulk.remove", data)
        response = response.json()
        if len(response["missing"]) > 0:
            logger.warning(
                f"Failed to remove items from Entities Collection. IDs: {response['missing']}. "
            )
        return response["items"]



import functools
from typing import Dict

import numpy as np

from supervisely.annotation.label import Label
from supervisely.decorators.inference import (
    _process_image_path,
    _scale_ann_to_original_size,
    process_image_sliding_window,
)
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.rectangle import Rectangle
from supervisely.imaging import image as sly_image
from supervisely.io.fs import silent_remove
from supervisely.nn.inference.semantic_segmentation.semantic_segmentation import (
    SemanticSegmentation,
)
from supervisely.nn.prediction_dto import PredictionMask
from supervisely.sly_logger import logger


class SalientObjectSegmentation(SemanticSegmentation):
    def get_info(self) -> dict:
        info = super().get_info()
        info["task type"] = "salient object segmentation"
        return info

    def _create_label(self, dto: PredictionMask):
        geometry = Bitmap(dto.mask, extra_validation=False)
        obj_class = self.model_meta.get_obj_class(dto.class_name)
        if not dto.mask.any():  # skip empty masks
            logger.debug(f"Mask is empty and will be slipped")
            return None
        label = Label(geometry, obj_class)
        return [label]

    # pylint: disable=no-self-argument
    def process_image_crop(func):
        """
        Decorator for processing annotation labels before and after inference.
        Crops input image before inference if kwargs['state']['rectangle_crop'] provided
        and then scales annotation back to original image size.
        Keyword arguments:
        :param image_np: Image in numpy.ndarray format (use image_path or image_np, not both)
        :type image_np: numpy.ndarray
        :param image_path: Path to image (use image_path or image_np, not both)
        :type image_path: str
        :raises: :class:`ValueError`, if image_np or image_path invalid or not provided
        :return: Annotation in json format
        :rtype: :class:`dict`
        """

        # function for bounding boxes padding
        def bbox_padding(rectangle, padding):
            padding = padding.strip()  # remove blank spaces
            if padding.endswith("px"):
                padding = int(padding[:-2])
                format = "pixels"
            elif padding.endswith("%"):
                padding = int(padding[:-1])
                padding = round(padding / 100, 2)  # from % to float
                format = "percentages"
            else:
                raise ValueError(
                    "Unsupported padding unit: only pixels (e.g. 10px) and percentages (e.g. 10%) are supported"
                )
            if padding < 0:
                padding = 0
            left, right, top, bottom = (
                rectangle.left,
                rectangle.right,
                rectangle.top,
                rectangle.bottom,
            )
            if format == "pixels":
                pad_left = left - padding
                pad_right = right + padding
                pad_top = top - padding
                pad_bottom = bottom + padding
            elif format == "percentages":
                width, height = rectangle.width, rectangle.height
                width_padding = int(width * padding)
                height_padding = int(height * padding)
                pad_left = left - width_padding
                pad_right = right + width_padding
                pad_top = top - height_padding
                pad_bottom = bottom + height_padding
            else:
                raise ValueError("Unsupported padding format")
            return Rectangle(pad_top, pad_left, pad_bottom, pad_right)

        # function for padded bounding boxes processing
        def process_padded_bbox(image, padded_bbox):
            img_height, img_width = image.shape[:2]
            x_min, y_min, x_max, y_max = 0, 0, img_width, img_height
            box_x_min, box_y_min, box_x_max, box_y_max = (
                padded_bbox.left,
                padded_bbox.top,
                padded_bbox.right,
                padded_bbox.bottom,
            )
            if box_x_min <= x_min:
                box_x_min = x_min + 1
            if box_y_min <= y_min:
                box_y_min = y_min + 1
            if box_x_max >= x_max:
                box_x_max = x_max - 1
            if box_y_max >= y_max:
                box_y_max = y_max - 1
            return Rectangle(box_y_min, box_x_min, box_y_max, box_x_max)

        @functools.wraps(func)
        def wrapper_inference(*args, **kwargs):
            # pylint: disable=not-callable
            settings = kwargs["settings"]
            rectangle_json = settings.get("rectangle")

            if rectangle_json is None:
                ann = func(*args, **kwargs)
                return ann

            rectangle = Rectangle.from_json(rectangle_json)
            padding = settings.get("bbox_padding")
            original_rectangle = None
            if padding is not None:
                original_rectangle = rectangle
                rectangle = bbox_padding(rectangle, padding)
                if "image_np" in kwargs.keys():
                    image_np = kwargs["image_np"]
                elif "image_path" in kwargs.keys():
                    image_path = kwargs["image_path"]
                    image_np = sly_image.read(image_path)
                else:
                    raise ValueError("image_np or image_path not provided!")
                rectangle = process_padded_bbox(image_np, rectangle)

            if "image_np" in kwargs.keys():
                image_np = kwargs["image_np"]
                if not isinstance(image_np, np.ndarray):
                    raise ValueError("Invalid input. Image path must be numpy.ndarray")
                original_image_size = image_np.shape[:2]
                image_crop_np = sly_image.crop(image_np, rectangle)
                kwargs["image_np"] = image_crop_np
                ann = func(*args, **kwargs)
                ann = _scale_ann_to_original_size(ann, original_image_size, rectangle)
                if padding:
                    ann = ann.crop_labels(
                        original_rectangle
                    )  # crop labels to avoid overlapping masks
            elif "image_path" in kwargs.keys():
                image_path = kwargs["image_path"]
                if not isinstance(image_path, str):
                    raise ValueError("Invalid input. Image path must be str")
                image_crop_path, original_image_size = _process_image_path(image_path, rectangle)
                kwargs["image_path"] = image_crop_path
                ann = func(*args, **kwargs)
                ann = _scale_ann_to_original_size(ann, original_image_size, rectangle)
                silent_remove(image_crop_path)
                if padding:
                    ann = ann.crop_labels(
                        original_rectangle
                    )  # crop labels to avoid overlapping masks
            else:
                raise ValueError("image_np or image_path not provided!")

            return ann

        return wrapper_inference

    @process_image_sliding_window
    @process_image_crop
    def _inference_image_path(
        self,
        image_path: str,
        settings: Dict,
        data_to_return: Dict,  # for decorators
    ):
        inference_mode = settings.get("inference_mode", "full_image")
        logger.debug(
            "Inferring image_path:", extra={"inference_mode": inference_mode, "path": image_path}
        )

        if inference_mode == "sliding_window" and settings["sliding_window_mode"] == "advanced":
            predictions = self.predict_raw(image_path=image_path, settings=settings)
        else:
            predictions = self.predict(image_path=image_path, settings=settings)
        ann = self._predictions_to_annotation(
            image_path, predictions, settings.get("classes", None)
        )

        logger.debug(
            f"Inferring image_path done. pred_annotation:",
            extra=dict(w=ann.img_size[1], h=ann.img_size[0], n_labels=len(ann.labels)),
        )
        return ann



from dataclasses import dataclass

import psutil


class ModelSource:
    PRETRAINED = "Pretrained models"
    CUSTOM = "Custom models"


class RuntimeType:
    PYTORCH = "PyTorch"
    ONNXRUNTIME = "ONNXRuntime"
    TENSORRT = "TensorRT"


class ModelPrecision:
    FP32 = "FP32"
    FP16 = "FP16"
    INT8 = "INT8"


@dataclass
class CheckpointInfo:
    """
    :param checkpoint_name: A name for model weights, e.g., "YOLOv8n COCO (best)".
    :param model_name: The name of a model for which the weights are applicable, e.g., "YOLOv8n".
    :param architecture: Collection for a set of models, e.g., "YOLOv8".
    :param checkpoint_url: URL to download the model weights.
    :param custom_checkpoint_path: Path in Team Files to the weights.
    :param model_source: Source of the model, either "Pretrained models" or "Custom models".
    """

    checkpoint_name: str = None
    model_name: str = None
    architecture: str = None
    checkpoint_url: str = None
    custom_checkpoint_path: str = None
    model_source: str = None


@dataclass
class DeployInfo:
    checkpoint_name: str
    model_name: str
    architecture: str
    checkpoint_url: str
    custom_checkpoint_path: str
    model_source: str
    task_type: str
    device: str
    runtime: str
    model_precision: str
    hardware: str
    deploy_params: dict


def _get_model_name(model_info: dict):
    name = model_info.get("model_name")
    if not name:
        name = model_info.get("meta", {}).get("model_name")
    if not name:
        name = model_info.get("Model")
    if not name:
        raise ValueError("Model name not found not in model_info nor in meta.")
    return name


def get_ram_usage():
    memory = psutil.virtual_memory()
    return memory.used, memory.total


def get_gpu_usage(device: str = None):
    if device == "cpu":
        return None, None
    try:
        import torch
    except Exception as e:
        from supervisely import logger
        logger.warning(f"Cannot import torch. Install PyTorch to get GPU usage info. Error: {e}")
        return None, None
    if not torch.cuda.is_available():
        return None, None
    gpu_index = None
    if device is None or device in ["", "auto", "cuda"]:
        gpu_index = torch.cuda.current_device()
    elif isinstance(device, int):
        gpu_index = device
    elif device.startswith("cuda:"):
        try:
            gpu_index = int(device.split(":")[-1])
        except ValueError:
            return None, None
    else:
        for i in range(torch.cuda.device_count()):
            if device == torch.cuda.get_device_name(i):
                gpu_index = i
                break
        if gpu_index is None:
            return None, None
    if gpu_index is None or gpu_index > torch.cuda.device_count() or gpu_index < 0:
        return None, None
    allocated = torch.cuda.memory_allocated(gpu_index)
    total = torch.cuda.get_device_properties(gpu_index).total_memory
    return allocated, total



# coding: utf-8
import json
import os
from typing import Dict, Optional

import aiofiles
import jsonschema


class JsonSerializable:
    def to_json(self):
        """Serialize to JSON-compatible dict.
        :return: dict
        """
        raise NotImplementedError()

    @classmethod
    def from_json(cls, data):
        """
        Deserialize from a JSON-compatible dict
        :param data: JSON-compatible dict
        :return: Parsed object
        """
        raise NotImplementedError()


def load_json_file(filename: str) -> Dict:
    """
    Decoding data from json file with given filename.

    :param filename: Target file path.
    :type filename: str
    :raises IsADirectoryError: If given filename is a directory, not a file.
    :raises FileNotFoundError: If file with given filename was not found.
    :raises RuntimeError: If can not decode json file.
    :returns: Json format as a dict
    :rtype: :class:`dict`
    :Usage example:

     .. code-block:: python

        from supervisely.io.json import load_json_file
        json_example = load_json_file('/home/admin/work/projects/examples/ann.json')
        print(json_example)
        # Output: {
        #     "description": "",
        #     "tags": [],
        #     "size": {
        #         "height": 800,
        #         "width": 1067
        #     },
        #     "objects": [
        #         {
        #             "id": 619053179,
        #             "classId": 2791451,
        #             "description": "",
        #             "geometryType": "bitmap",
        #             "labelerLogin": "alexxx",
        #             "createdAt": "2021-02-10T08:36:33.898Z",
        #             "updatedAt": "2021-02-10T08:39:24.828Z",
        #             "tags": [],
        #             "classTitle": "lemon",
        #             "bitmap": {
        #                 "data": "eJwBZgOZ/IlQTkcNChoKAAAADUlIR...AiRp9+EwAAAABJRU5ErkJgggn2cM0=",
        #                 "origin": [
        #                     531,
        #                     120
        #                 ]
        #             }
        #         },
        #         {
        #             "id": 619053347,
        #             "classId": 2791453,
        #             "description": "",
        #             "geometryType": "rectangle",
        #             "labelerLogin": "alexxx",
        #             "createdAt": "2021-02-10T08:39:08.369Z",
        #             "updatedAt": "2021-02-10T08:39:24.828Z",
        #             "tags": [],
        #             "classTitle": "kiwi",
        #             "points": {
        #                 "exterior": [
        #                     [
        #                         764,
        #                         387
        #                     ],
        #                     [
        #                         967,
        #                         608
        #                     ]
        #                 ],
        #                 "interior": []
        #             }
        #         },
        #         {
        #             "id": 619053355,
        #             "classId": 2791453,
        #             "description": "",
        #             "geometryType": "rectangle",
        #             "labelerLogin": "alexxx",
        #             "createdAt": "2021-02-10T08:39:16.938Z",
        #             "updatedAt": "2021-02-10T08:39:24.828Z",
        #             "tags": [],
        #             "classTitle": "kiwi",
        #             "points": {
        #                 "exterior": [
        #                     [
        #                         477,
        #                         543
        #                     ],
        #                     [
        #                         647,
        #                         713
        #                     ]
        #                 ],
        #                 "interior": []
        #             }
        #         }
        #     ]
        # }
    """
    if os.path.isdir(filename):
        raise IsADirectoryError(f"The path {filename} is a directory, not a file.")
    elif not os.path.isfile(filename):
        raise FileNotFoundError(f"File with path {filename} was not found.")
    try:
        with open(filename, encoding="utf-8") as fin:
            return json.load(fin)
    except json.decoder.JSONDecodeError as e:
        raise RuntimeError(
            f"Can not decode json file with path {filename}: {e.msg} at "
            f"line number: {e.lineno}, column: {e.colno}, position: {e.pos}. "
            f"Document: {e.doc}"
        )


def dump_json_file(data: Dict, filename: str, indent: Optional[int] = 4) -> None:
    """
    Write given data in json format in file with given name.

    :param data: Data in json format as a dict.
    :type data: dict
    :param filename: Target file path to write data.
    :type filename: str
    :param indent: Json array elements and object members will be pretty-printed with that indent level.
    :type indent: int, optional
    :returns: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        from supervisely.io.json import dump_json_file
        data = {1: 'example'}
        dump_json_file(data, '/home/admin/work/projects/examples/1.json')
    """
    with open(filename, "w") as fout:
        json.dump(data, fout, indent=indent)


def flatten_json(data: Dict, sep: Optional[str] = ".") -> Dict:
    """
    Normalize semi-structured JSON data into a flat table.

    :param data: Data in json format as a dict.
    :type data: dict
    :param sep: Nested records will generate names separated by sep.
    :type sep: str, optional
    :returns: Dict
    :rtype: :class:`dict`
    """
    import pandas as pd

    df = pd.json_normalize(data, sep=sep)
    return df.to_dict(orient="records")[0]


def modify_keys(
    data: Dict, prefix: Optional[str] = None, suffix: Optional[str] = None
) -> Dict[str, str]:
    """
    Add prefix and suffix to keys of given dict.

    :param data: Data in json format as a dict.
    :type data: dict
    :param prefix: Prefix which will be added to dict.
    :type prefix: str, optional
    :param suffix: Suffix which will be added to dict.
    :type suffix: str, optional
    :returns: New dict with prefix and suffix in keys
    :rtype: :class:`dict`
    :Usage example:

     .. code-block:: python

        from supervisely.io.json import modify_keys
        data = {'1': 'example', '3': 4}
        new_data = modify_keys(data, prefix='pr_', suffix='_su')
        print(new_data)
        # Output: {'pr_1_su': 'example', 'pr_3_su': 4}
    """

    def _modify(k):
        res = k
        if prefix is not None:
            res = prefix + res
        if suffix is not None:
            res += suffix
        return res

    return {_modify(k): v for k, v in data.items()}


def validate_json(data: Dict, schema: Dict, raise_error: bool = False) -> bool:
    """
    Validate json data.

    :param data: Data in json format as a dict.
    :type data: dict
    :param schema: Schema in json format as a dict.
    :type schema: dict
    :param raise_error: If True, raise an error if data is invalid.
    :type raise_error: bool, optional
    :returns: True if data is valid, False otherwise.
    :rtype: :class:`bool`
    """
    try:
        jsonschema.validate(instance=data, schema=schema)
        return True
    except jsonschema.exceptions.ValidationError as err:
        if raise_error:
            raise ValueError("JSON data is invalid. See error message for more details.") from err
        return False


async def dump_json_file_async(data: Dict, filename: str, indent: Optional[int] = 4) -> None:
    """
    Write given data in json format in file with given name asynchronously.

    :param data: Data in json format as a dict.
    :type data: dict
    :param filename: Target file path to write data.
    :type filename: str
    :param indent: Json array elements and object members will be pretty-printed with that indent level.
    :type indent: int, optional
    :returns: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        import supervisely as sly
        from supervisely._utils import run_coroutine

        data = {1: 'example'}

        coroutine = sly.json.dump_json_file_async(data, '/home/admin/work/projects/examples/1.json')
        run_coroutine(coroutine)
    """
    async with aiofiles.open(filename, "w") as fout:
        await fout.write(json.dumps(data, indent=indent))



from typing import Any, Dict, Optional, Tuple, Union

from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget


class BindedInputNumber(Widget):
    """BindedInputNumber widget in Supervisely is a user interface element that allows users
    to input two numerical values and customize their behavior using the proportional, min, and max properties.

    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/input/bindedinputnumber>`_
        (including screenshots and examples).

    :param width: Width of the image in pixels.
    :type width: Optional[int]
    :param height: Height of the image in pixels.
    :type height: Optional[int]
    :param min: Minimum value of the input.
    :type min: Optional[int]
    :param max: Maximum value of the input.
    :type max: Optional[int]
    :param proportional: If True, the input will be proportional.
    :type proportional: Optional[bool]
    :param widget_id: Unique widget identifier.
    :type widget_id: str

    :Usage example:
    .. code-block:: python
        from supervisely.app.widgets import Badge

        binded_input_number = BindedInputNumber(width=150, height=150, min=1, max=100, proportional=False)
    """

    def __init__(
        self,
        width: Optional[int] = 256,
        height: Optional[int] = 256,
        min: Optional[int] = 1,
        max: Optional[int] = 10000,
        proportional: Optional[bool] = False,
        widget_id: Optional[str] = None,
    ):
        self._width = width
        self._height = height
        self._min = min
        self._max = max
        self._proportional = proportional
        self._disabled = False

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict[str, Any]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - options: Dictionary with the following fields:
                - proportions: Dictionary with the following fields:
                    - width: Width of the image in pixels.
                    - height: Height of the image in pixels.
            - disabled: If True, the widget will be disabled.
        """
        return {
            "options": {"proportions": {"width": self._width, "height": self._height}},
            "disabled": self._disabled,
        }

    def get_json_state(self) -> Dict[str, Union[int, bool]]:
        """Returns dictionary with widget state.

        Dictionary contains the following fields:
               - value: Dictionary with the following fields:
                   - min: Minimum value of the input.
                   - max: Maximum value of the input.
                   - width: Width of the image in pixels.
                   - height: Height of the image in pixels.
                   - proportional: If True, the input will be proportional.
        """
        return {
            "value": {
                "min": self._min,
                "max": self._max,
                "width": self._width,
                "height": self._height,
                "proportional": self._proportional,
            }
        }

    @property
    def value(self) -> Tuple[int, int]:
        """Returns the width and height of the image.

        :return: Tuple with the width and height of the image.
        :rtype: Tuple[int, int]
        """
        return self._width, self._height

    @value.setter
    def value(self, width: int, height: int) -> None:
        """Sets the width and height of the image.

        :param width: Width of the image in pixels.
        :type width: int
        :param height: Height of the image in pixels.
        :type height: int
        """
        self._width = width
        self._height = height
        StateJson()[self.widget_id]["width"] = self._width
        StateJson()[self.widget_id]["height"] = self._height
        StateJson().send_changes()

    def get_value(self) -> Tuple[int, int]:
        """Returns the width and height of the image.

        :return: Tuple with the width and height of the image.
        :rtype: Tuple[int, int]
        """
        width = StateJson()[self.widget_id]["value"]["width"]
        height = StateJson()[self.widget_id]["value"]["height"]
        return width, height

    @property
    def proportional(self) -> bool:
        """Returns True if the input is proportional, False otherwise."""
        return self._proportional

    @property
    def min(self) -> int:
        """Returns the minimum value of the input."""
        return self._min

    @property
    def max(self) -> int:
        """Returns the maximum value of the input."""
        return self._max

    @proportional.setter
    def proportional(self, value) -> None:
        """Sets the proportional value.

        :param value: If True, the input will be proportional.
        :type value: bool

        :Usage example:
        .. code-block:: python
            binded_input_number.proportional = True
        """
        self._proportional = value
        DataJson()[self.widget_id]["proportional"] = self._proportional
        DataJson().send_changes()

    @min.setter
    def min(self, value) -> None:
        """Sets the minimum value of the input.

        :param value: Minimum value of the input.
        :type value: int
        """
        self._min = value
        DataJson()[self.widget_id]["min"] = self._min
        DataJson().send_changes()

    @max.setter
    def max(self, value) -> None:
        """Sets the maximum value of the input.

        :param value: Maximum value of the input.
        :type value: int
        """
        self._max = value
        DataJson()[self.widget_id]["max"] = self._max
        DataJson().send_changes()

    def disable(self) -> None:
        """Disables the widget in the UI."""
        self._disabled = True
        DataJson()[self.widget_id]["disabled"] = self._disabled
        DataJson().send_changes()

    def enable(self) -> None:
        """Enables the widget in the UI."""
        self._disabled = False
        DataJson()[self.widget_id]["disabled"] = self._disabled
        DataJson().send_changes()
